package org.openxma.dsl.generator.mojo;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.mwe.core.resources.CachingResourceLoaderImpl;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderFactory;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderImpl;
import org.eclipse.xtext.ISetup;
import org.openxma.dsl.generator.DefaultGeneratorConfiguration;
import org.openxma.dsl.generator.GeneratorConfiguration;

import com.google.inject.Injector;
import com.google.inject.Provider;

import at.spardat.xma.guidesign.util.ClassLoaderFactory;

public abstract class AbstractGeneratorMojo extends AbstractMojo {

    /**
     * The project itself. This parameter is set by maven.
     * 
     * @parameter expression="${project}"
     * @required
     */
    protected MavenProject project;
    /**
     * @parameter default-value="${project.artifacts}"
     * @required
     * @readonly
     */
    private Set<Artifact> dependencies;

    /**
     * Base directory of the project.
     * 
     * @parameter expression="${basedir}"
     */
    protected File baseDirectory;

    /**
     * @required
     * @parameter expression="${project.build.sourceDirectory}"
     */
    protected File sourceDirectory;
    /**
     * @required
     * @parameter expression="${basedir}/src/main/resources"
     */
    protected File resourceDirectory;
    /**
     * @parameter expression="${project.build.testSourceDirectory}"
     */
    protected File testSourceDirectory;
    /**
     * @required
     * @parameter expression="${basedir}/src/generated/java"
     */
    protected File generatedSourceDirectory;
    /**
     * @required
     * @parameter expression="${basedir}/src/generated/resources"
     */
    protected File generatedResourceDirectory;

    /**
     * @parameter expression="${basedir}/src/test/generated/java"
     */
    protected File generatedTestSourceDirectory;

    /**
     * The location of the generator properties file.
     * 
     * @parameter alias="properties"
     */
    protected File generatorProperties;
    /**
     * @required
     * @parameter default-value="${basedir}/src/main/model"
     */
    protected String modelFolder;
    /**
     * parameter to control if generated source directories should be deleted prior to generating new files
     * 
     * @parameter default-value="false"
     */
    protected boolean cleanDirectory;

    private List<URL> classLoaderUrls = new ArrayList<URL>();

    public void execute() throws MojoExecutionException, MojoFailureException {
        if (!Boolean.getBoolean("skipGenerator")) {
            System.setProperty("org.eclipse.emf.ecore.EPackage.Registry.INSTANCE",
                    "org.eclipse.emf.ecore.impl.EPackageRegistryImpl");
            doExecute();
        }
    }

    public abstract void doExecute() throws MojoExecutionException;

    protected DefaultGeneratorConfiguration createGeneratorConfiguration(String fileExtension)
            throws MojoExecutionException {
        Properties properties = loadProperties();
        DefaultGeneratorConfiguration generatorConfiguration = new DefaultGeneratorConfiguration(properties);
        generatorConfiguration.setProjectFolder(URI.createURI(baseDirectory.toURI().toString()));
        generatorConfiguration.setFileExtension(fileExtension);
        generatorConfiguration.setGenResourceFolder(URI.createURI(generatedResourceDirectory.toURI().toString()));
        generatorConfiguration.setGenSourceFolder(URI.createURI(generatedSourceDirectory.toURI().toString()));
        generatorConfiguration.setGenTestSourceFolder(URI.createURI(generatedTestSourceDirectory.toURI().toString()));
        generatorConfiguration.setResourceFolder(URI.createURI(resourceDirectory.toURI().toString()));
        generatorConfiguration.setSourceFolder(URI.createURI(sourceDirectory.toURI().toString()));
        generatorConfiguration.setTestSourceFolder(URI.createURI(testSourceDirectory.toURI().toString()));
        GeneratorConfiguration.INSTANCE.set(generatorConfiguration);
        return generatorConfiguration;
    }

    protected Properties loadProperties() throws MojoExecutionException {
        Properties properties = new Properties();
        if (generatorProperties != null) {
            getLog().info("load generator properties from '" + generatorProperties.getPath() + "'");
            try {
                properties.load(new FileInputStream(generatorProperties));
            } catch (FileNotFoundException e) {
                throw new MojoExecutionException(e.getMessage(), e);
            } catch (IOException e) {
                throw new MojoExecutionException(e.getMessage(), e);
            }
        }
        return properties;
    }

    protected void extendCurrentClassloader(ClassLoader currentClassLoader) {
        if (classLoaderUrls.isEmpty()) {
            classLoaderUrls.add(toURL(resolvePath(new File(project.getBuild().getSourceDirectory())), false));
            List<?> resources = project.getBuild().getResources();
            if (resources != null) {
                for (int i = 0; i < resources.size(); i++) {
                    org.apache.maven.model.Resource resource = (org.apache.maven.model.Resource) resources.get(i);
                    File directory = resolvePath(new File(resource.getDirectory()));
                    classLoaderUrls.add(toURL(directory, false));
                }
            }
            for (Artifact artifact : dependencies) {
                URL url = toURL(artifact.getFile(), false);
                classLoaderUrls.add(url);
            }
            // TODO: consider active maven profiles
            if (classLoaderUrls.size() > 0) {
                File outputDirectory = resolvePath(new File(project.getBuild().getOutputDirectory()));
                classLoaderUrls.add(toURL(outputDirectory, false));
                URL[] urls = classLoaderUrls.toArray(new URL[classLoaderUrls.size()]);
                ClassLoader urlClassLoader = new URLClassLoader(urls, currentClassLoader);
                ResourceLoaderFactory.setCurrentThreadResourceLoader(new CachingResourceLoaderImpl(
                        new ResourceLoaderImpl(urlClassLoader)));
                // required for guidesigner!
                ClassLoaderFactory.setProjectClassLoader(urlClassLoader);
                // required for backwards compat. with Pom2Xma.mwe can be
                // removed if we change and remove the classpathUri to 0fileUri
                Thread.currentThread().setContextClassLoader(urlClassLoader);
                String separator = System.getProperty("path.separator");
                StringBuilder stringBuilder = new StringBuilder();
                for (URL url : urls) {
                    File file = new File(url.getPath());
                    if (file.exists() && file.isDirectory()
                            || (file.getName().endsWith(".jar") || file.getName().endsWith(".zip"))) {
                        stringBuilder.append(url.getPath());
                        stringBuilder.append(separator);
                    }
                }
                System.setProperty("java.class.path", stringBuilder.toString());
            }
        }
    }

    protected ISetup createCachedSetup(final ISetup setup) {
        return new ISetup() {

            @SuppressWarnings("unchecked")
            public Injector createInjectorAndDoEMFRegistration() {
                Injector injector = (Injector) getPluginContext().get(setup.getClass());
                if (injector == null) {
                    injector = setup.createInjectorAndDoEMFRegistration();
                    getPluginContext().put(setup.getClass(), injector);
                }
                return injector;
            }

        };
    }
    
    protected Provider<ResourceSet> createSharedResourceSet(final ISetup setup) {
//        return new Provider<ResourceSet>() {
//
//            @SuppressWarnings("unchecked")
//            public ResourceSet get() {
//                ResourceSet resourceSet = (ResourceSet) getPluginContext().get("_shared_resourceSet");
//                if (resourceSet == null) {
//                    resourceSet = createCachedSetup(setup).createInjectorAndDoEMFRegistration().getInstance(
//                            ResourceSet.class);
//                    getPluginContext().put("_shared_resourceSet", resourceSet);
//                }
//                return resourceSet;
//            }
//        };
        return null;
    }
    
    private URL toURL(File path, boolean trailingSlash) {
        URL url = null;
        try {
            url = new URL(path.toURI().toString() + ((trailingSlash) ? "/" : ""));
        } catch (Exception e) {
            getLog().error("Could not resolve \"" + path.getPath() + "\".", e);
        }
        return url;
    }

    private File resolvePath(File path) {
        if (!path.isAbsolute()) {
            path = new File(project.getBasedir(), path.getPath());
        }
        if (path.isDirectory()) {
            String name = path.getName();
            if (!name.endsWith(File.separator)) {
                path = new File(path.getPath() + File.separator);
            }
        }
        return path;
    }

    protected String join(Object[] array, String separator, int startIndex, int endIndex) {
        if (array == null) {
            return null;
        }
        if (separator == null) {
            separator = "";
        }
        int bufSize = (endIndex - startIndex);
        if (bufSize <= 0) {
            return "";
        }
        bufSize *= ((array[startIndex] == null ? 16 : array[startIndex].toString().length()) + separator.length());
        StringBuffer buf = new StringBuffer(bufSize);
        for (int i = startIndex; i < endIndex; i++) {
            if (i > startIndex) {
                buf.append(separator);
            }
            if (array[i] != null) {
                buf.append(array[i]);
            }
        }
        return buf.toString();
    }

}
