package org.openxma.dsl.generator.mojo;

import static com.google.common.collect.Collections2.transform;
import static com.google.common.collect.Sets.newHashSet;
import static java.util.Arrays.asList;
import static org.eclipse.emf.common.util.URI.createFileURI;
import static org.openxma.dsl.common.DslConstants.OPENXMA_POM_FILE_EXTENSION;

import java.io.File;
import java.util.Collection;
import java.util.Properties;
import java.util.Set;

import javax.annotation.Nullable;

import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.shared.model.fileset.FileSet;
import org.apache.maven.shared.model.fileset.util.FileSetManager;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.mwe.core.WorkflowContextDefaultImpl;
import org.eclipse.emf.mwe.core.issues.IssuesImpl;
import org.eclipse.emf.mwe.core.monitor.NullProgressMonitor;
import org.eclipse.emf.mwe.utils.StandaloneSetup;
import org.eclipse.xtext.mwe.NameBasedFilter;
import org.eclipse.xtext.mwe.SlotEntry;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.dom.DomDslStandaloneSetup;
import org.openxma.dsl.generator.DefaultGeneratorConfiguration;
import org.openxma.dsl.generator.DefaultGeneratorFactory;
import org.openxma.dsl.generator.Generator;
import org.openxma.dsl.generator.GeneratorConstants;
import org.openxma.dsl.generator.component.DslReader;
import org.openxma.dsl.pom.PomDslStandaloneSetup;
import org.openxma.dsl.pom.PomPackage;

import com.google.common.base.Function;

/**
 * Maven plugin used to execute the openXMA Pom generator workflow
 * 
 * @goal generate-pom
 * @requiresDependencyResolution compile
 * @phase generate-sources
 */
public class PomGeneratorMojo extends AbstractGeneratorMojo {
    /**
     * @required
     * @parameter
     */
    protected FileSet componentFileSet;
    /**
     * @parameter default-value="/src/main/resources/guidesign.properties"
     * @required
     * @readonly
     */
    private String guiDesignProperties;

    public void doExecute() throws MojoExecutionException {
        check();
        setup();
        cleanDirectory();
        Collection<Model> models = loadModels();
        generate(models);
    }

    protected void check() throws MojoExecutionException {
        if (componentFileSet == null || componentFileSet.getIncludes() == null
                || componentFileSet.getIncludes().size() == 0) {
            throw new MojoExecutionException(
                    "No files specified. User either the 'file' attribute or a 'fileset' element.");
        }
    }

    protected void cleanDirectory() {
    }

    protected void setup() {
        extendCurrentClassloader(getClass().getClassLoader());
        StandaloneSetup standaloneSetup = new StandaloneSetup();
        standaloneSetup.addRegisterGeneratedEPackage(PomPackage.class.getName());
    }

    protected void generate(Collection<Model> models) throws MojoExecutionException {
        Generator generator = createGenerator();
        if (models != null) {
            FileSetManager fileSetManager = new FileSetManager();
            String[] includedFiles = fileSetManager.getIncludedFiles(componentFileSet);

            Set<URI> selected = newHashSet(transform(asList(includedFiles), new Function<String, URI>() {

                public URI apply(@Nullable String input) {
                    File directory = new File(componentFileSet.getDirectory());
                    return createFileURI(new File(directory, input).toString());
                }
            }));
            for (Model model : models) {
                Resource resource = model.eResource();
                URI uri = resource.getURI();
                if (uri.isArchive()) {
                    continue;
                }
                if (selected.contains(uri)) {
                    generator.generate(resource, new NullProgressMonitor());
                }
            }
        }
    }

    protected Generator createGenerator() throws MojoExecutionException {
        DefaultGeneratorConfiguration generatorConfiguration = createGeneratorConfiguration(OPENXMA_POM_FILE_EXTENSION);
        Generator generator = new DefaultGeneratorFactory().createAndInit(generatorConfiguration);
        return generator;
    }

    @SuppressWarnings("unchecked")
    protected Collection<Model> loadModels() {
        PomDslStandaloneSetup setup = new PomDslStandaloneSetup();
        DslReader reader = new DslReader();
        reader.setUseJavaClassPath(true);
        reader.addRegister(createCachedSetup(new DomDslStandaloneSetup()));
        reader.addRegister(createCachedSetup(setup));
        reader.addPath(modelFolder);
        reader.setResourceSetProvider(createSharedResourceSet(setup));
        SlotEntry slotEntry = new SlotEntry();
        slotEntry.setSlot("model");
        slotEntry.setType(Model.class.getSimpleName());
        reader.addLoad(slotEntry);
        NameBasedFilter nameBasedFilter = new NameBasedFilter();
        nameBasedFilter.setRegularExpression("^(?:(?!classes|bin).)*\\.(pml|xma|dml|xmadsl)$");
        reader.setUriFilter(nameBasedFilter);
        WorkflowContextDefaultImpl workflowContext = new WorkflowContextDefaultImpl();
        reader.invoke(workflowContext, new NullProgressMonitor(), new IssuesImpl());
        return (Collection<Model>) workflowContext.get("model");
    }

    @Override
    protected Properties loadProperties() throws MojoExecutionException {
        Properties properties = super.loadProperties();
        properties.setProperty(GeneratorConstants.GUIDESIGNER_PROPERTIES_FILE, guiDesignProperties);
        return properties;
    }

}
